{-

An interpreter for a simple functional language
that can easily represent the exponentiation function
and functions on pairs and lists.

-}

module Eval where

import Prelude hiding (exp)
import Data.Maybe

-- Abstract syntax

-- BEGIN_PROG
type Prog = ([FDef], Expr)
-- END_PROG

-- BEGIN_FDEF
type FDef = (String,([String],Expr))
-- END_FDEF

-- BEGIN_EXP_VAL
data Val
 = IVal  { getInt :: Int }
 | BVal { getBool :: Bool }
-- END_EXP_VAL
 | Char { getChar :: Char }
 | Pair { getPair :: (Val,Val) }
 | List { getList :: [Val] }
 | Error String

-- BEGIN_EXP_TYPE
data Expr
 = Const { getVal :: Val }
 | Var String
 | Apply String [Expr]
 | Prim Op [Expr]
 | If Expr Expr Expr
-- END_EXP_TYPE
 | Let String Expr Expr

instance Show Expr where
 show (Const v) = show v
 show (Var s) = s
 show (Apply fun args) = fun ++ show args
 show (Prim op args) = show op ++ " " ++ show args
 show (If c a b) = "if " ++ show c ++ " then " ++ show a ++ " else " ++ show b
 show (Let s e b) = "let " ++ s ++ "=" ++ show e ++ "in\n" ++ show b

instance Show Val
 where
  show (Error s) = "<" ++ s ++ ">"
  show (IVal i) = show i
  show (BVal b) = show b
  show (Char c) = show c
  show (Pair p) = show p
  show (List l) = show l

-- BEGIN_EXP_OP
data Op = Equal | Add | Sub | Mul
-- END_EXP_OP
 | MkPair
 | Cons -- binary operators
 | Fst
 | Snd
 | Null
 | Head
 | Tail -- unary operators

instance Show Op where
  show Equal = "=="
  show Add = "+"
  show Sub = "-"
  show Mul = "*"
  show MkPair = ","
  show Cons = ":"
  show Fst = "FST"
  show Snd = "SND"
  show Null = "NULL"
  show Head = "HEAD"
  show Tail = "TAIL"

-- Convenience

nil = Const (List [])


-- Evaluator

-- Environment
-- BEGIN_ENV
type Env = [(String, Val)]
-- END_ENV

-- BEGIN_EVAL
eval :: Prog -> Val
eval (fdefs, main) = eval' main []

 where

  eval' :: Expr -> Env -> Val

  eval' (Const v) env = v
  
  eval' (Var s) env = 
    case lookup s env of
      Just v -> v
      Nothing -> error "undefined variable"
  
  eval' (Prim op es) env = 
    let rs = [ eval' e env | e <- es ] in
     prim op rs
  
  eval' (If e0 e1 e2) env =
    if getBool (eval' e0 env)
      then eval' e1 env
      else eval' e2 env
    
  eval' (Apply f es) env =
    eval' body env'
    where
      (ss, body) = fromJust (lookup f fdefs)
      env' = zip ss [ eval' e env | e <- es ]  
-- END_EVAL

  eval' (Let s e1 e2) env =
    eval' e2 env'
    where env' = (s, eval' e1 env):env

prim _ [e@(Error _)] = e
prim Fst [Pair p] = fst p
prim Snd [Pair p] = snd p
prim Null [List l] = BVal (null l)
prim Head [List (x:_)] = x
prim Tail [List (_:xs)] = List xs

prim _ [e@(Error _), _] = e
prim _ [_, e@(Error _)] = e
prim Equal [BVal b1, BVal b2] = BVal (b1 == b2)
prim Equal [Char c1, Char c2] = BVal (c1 == c2)
-- BEGIN_BASIC_PRIMS
prim Equal [IVal i1, IVal i2] = BVal (i1 == i2)
prim Add [IVal i1, IVal i2] = IVal (i1 + i2)
prim Sub [IVal i1, IVal i2] = IVal (i1 - i2)
prim Mul [IVal i1, IVal i2] = IVal (i1 * i2)
-- END_BASIC_PRIMS
prim MkPair [v1, v2] = Pair (v1, v2)
prim Cons [v, List l] = List (v:l)
prim uop vs = Error (show uop ++ " " ++ show vs)


-- Sample function exp

-- BEGIN_EXP_FUN
exp = (
 "exp", (["x","n"],
  If (Prim Equal [Var "n", Const (IVal 0)])
     (Const (IVal 1))  
     (Prim Mul 
       [Var "x",
        Apply "exp" [Var "x", Prim Sub [Var "n", Const (IVal 1)]]])))
-- END_EXP_FUN


-- Exp function with sample application

expProg = ([exp], Apply "exp" [Const (IVal 2), Const (IVal 3)])


-- Function environment with DFA simulators.
-- See Dfa2.hs for "run".
-- See Dfa3.hs for "run'".

dfaFenv = [
 ("run", (["s", "fs", "tf", "ls"],
    If (Prim Null [Var "ls"])
       (Apply "elem" [Var "s", Var "fs"])
       (Let "x" (Apply "lookup" [Var "s", Var "tf"])
        (Let "y" (Apply "lookup'" [Prim Head [Var "ls"], Var "x"])
         (If (Prim Null [Var "y"])
             (Const (BVal False))
             (Apply "run" [
                Prim Head [Var "y"],
                Var "fs",
                Var "tf",
                Prim Tail [Var "ls"]])))))),

 ("run'", (["s", "fs", "tf", "ls"],
    If (Prim Null [Var "ls"])
       (Apply "elem" [Var "s", Var "fs"])
       (Apply "search'n'run" [
          Apply "lookup" [Var "s", Var "tf"],
          Var "fs",
          Var "tf",
          Var "ls"]))),

 ("search'n'run", (["s'", "fs", "tf", "ls"],
    If (Prim Null [Var "s'"])
       (Const (BVal False))
       (If (Prim Equal [Prim Head [Var "ls"],
                      Prim Fst [Prim Head [Var "s'"]]])
           (Apply "run'" [
              Prim Snd [Prim Head [Var "s'"]],
              Var "fs",
              Var "tf",
              Prim Tail [Var "ls"]])
           (Apply "search'n'run" [
              Prim Tail [Var "s'"],
              Var "fs",
              Var "tf",
              Var "ls"])))),

 ("elem", (["x", "xs"], 
    If (Prim Null [Var "xs"])
       (Const (BVal False))
       (If (Prim Equal [Var "x", Prim Head [Var "xs"]])
           (Const (BVal True))
           (Apply "elem" [Var "x", Prim Tail [Var "xs"]])))),

 ("lookup", (["x", "xys"], 
    (If (Prim Equal [Var "x", Prim Fst [Prim Head [Var "xys"]]])
        (Prim Snd [Prim Head [Var "xys"]])
        (Apply "lookup" [Var "x", Prim Tail [Var "xys"]])))),

 ("lookup'", (["x", "xys"], 
    If (Prim Null [Var "xys"])
       nil
       (If (Prim Equal [Var "x", Prim Fst [Prim Head [Var "xys"]]])
           (Prim Cons [Prim Snd [Prim Head [Var "xys"]], nil])
           (Apply "lookup'" [Var "x", Prim Tail [Var "xys"]]))))

 ]

elemProg = (dfaFenv, Apply "elem" 
     [Const (IVal 1), Const (List [IVal 1])]) 

lookupProg = (dfaFenv, Apply "lookup" 
     [Const (IVal 2),
      Const (List [
        Pair (IVal 1, IVal 11),
        Pair (IVal 2, IVal 22),
        Pair (IVal 3, IVal 33)])]) 

-- Final states and transition function for an example

fs = List [IVal 2]
tf = List [Pair (IVal 1,List [Pair (Char 'a',IVal 2)]),
           Pair (IVal 2,List [Pair (Char 'b',IVal 1)])]

runProg ls = (dfaFenv, Apply "run"
     [Const (IVal 1),
      Const fs,
      Const tf,
      ls])

runProg' ls = (dfaFenv, Apply "run'"
     [Const (IVal 1),
      Const fs,
      Const tf,
      ls])
       

main
 = do
      print $ eval expProg
      print $ eval elemProg
      print $ eval lookupProg
      print $ eval (runProg (Const (List [Char 'a'])))
      print $ eval (runProg (Const (List [Char 'b'])))
      print $ eval (runProg (Const (List [Char 'a', Char 'b'])))
      print $ eval (runProg (Const (List [Char 'a', Char 'b', Char 'a'])))
      print $ eval (runProg' (Const (List [Char 'a'])))
      print $ eval (runProg' (Const (List [Char 'b'])))
      print $ eval (runProg' (Const (List [Char 'a', Char 'b'])))
      print $ eval (runProg' (Const (List [Char 'a', Char 'b', Char 'a'])))
