{-

A partial evaluator that aggregates the residual program. The simple
inlining-oriented predecessor (i.e., SimplePEval) is enhanced by a
state monad to aggregate the residual program as a function
environment.

-}

import Prelude hiding (exp)
import Data.Maybe
import Data.HashTable (hashString)
import Eval (Prog, FDef, Expr(..), Val(..), Op(..), Env, prim, exp, expProg, dfaFenv, fs, tf, elemProg, lookupProg, runProg)
import SimplePEval (isVal, expProg', expProg'')
import Control.Monad.State
-- import Data.Tuple (swap)

swap (x,y) = (y,x)

-- Partial evaluator

-- BEGIN_PEVAL_ZERO
peval :: Prog -> Prog
peval (fdefs, main) = swap (runState (peval' main []) [])

 where

  peval' :: Expr -> Env -> State [FDef] Expr 
-- END_PEVAL_ZERO

-- BEGIN_PEVAL_FIRST
  peval' (Const v) env = return (Const v)

  peval' (Var s) env = 
    case lookup s env of
      Just v  -> return (Const v)
      Nothing -> return (Var s)

  peval' (Prim op es) env = do
    rs <- mapM (flip peval' env) es
    if all isVal rs
     then return (Const (prim op (map getVal rs)))
     else return (Prim op rs)

  peval' (If e0 e1 e2) env = do 
    r0 <- peval' e0 env
    if isVal r0 then
      if getBool (getVal r0) 
       then peval' e1 env
       else peval' e2 env
     else do 
      r1 <- peval' e1 env
      r2 <- peval' e2 env
      return (If r0 r1 r2)
-- END_PEVAL_FIRST

  peval' (Let s e1 e2) env = do
    r1 <- peval' e1 env
    if isVal r1
      then peval' e2 ((s, getVal r1):env)
      else do
        r2 <- peval' e2 env
        return (Let s r1 r2)

-- BEGIN_PEVAL_APPLY
  peval' (Apply s es) env = do

    -- Look up function.
    let (ss, body) = fromJust (lookup s fdefs)

    -- Partially evaluate arguments.
    rs <- mapM (flip peval' env) es

    -- Determine static and dynamic arguments.
    let z = zip ss rs
    let sas = [ (s, getVal r) | (s, r) <- z, isVal r ]
    let das = [ (s,v) | (s,v) <- z, not (isVal v) ]

    if null das then 

      -- Inline completely static applications.
      peval' body sas 
  
      -- Otherwise these applications make their contexts dynamic.
     else do

        -- Fabricate name from static variables.
        let s' = s ++ show (hashString (show sas))

        -- Specialize each "name" just once.
        fdefs <- get
        when (isNothing (lookup s' fdefs)) (do

           -- Create placeholder for memoization.
           put (fdefs ++ [(s', undefined)]) 

           -- Partially evaluate function body.
           e' <- peval' body sas

           -- Replace placeholder by actual definition.
           modify (update (const (map fst das, e')) s'))

        -- Return application of specialized function.
        return (Apply s' (map snd das))
-- END_PEVAL_APPLY

-- Update a list that is supposed to be a map/dictionary

update :: Eq k => (v -> v) -> k -> [(k,v)] -> [(k,v)]
update f k ((k',v):kvs) = if k==k' then (k',f v):kvs else (k',v):update f k kvs


-- Specialized show for static arguments

myshowl [] = "[]"
myshowl (x:xs) = "[" ++ myshow1 x ++ myshowr xs ++ "]"
myshow1 (s,i) = "(" ++ s ++ "," ++ show i ++ ")"
myshowr [] = ""
myshowr (x:xs) = "," ++ myshow1 x ++ myshowr xs


-- Testing

runPe run = (dfaFenv, Apply run
     [Const (IVal 1),
      Const fs,
      Const tf,
      Var "ls"])
       
printBind (name, (args, body)) = do 
    putStr "  " 
    putStr name
    print args
    putStr "    "
    print $ body
    
test (env,exp) = do
  putStr "\n> "
  print $ exp
  sequence [ printBind bnd | bnd <- env ]
  putStr "* "
  let (env',exp') = peval (env,exp)
  print exp'
  sequence [ printBind bnd | bnd <- env' ]
  
main
 = do 
      test expProg
      test expProg'
      test expProg''
      test elemProg
      test lookupProg
      test (runProg (Const (List [Char 'a'])))
      test (runProg (Const (List [Char 'b'])))
      test (runProg (Const (List [Char 'a', Char 'b'])))
      test (runProg (Const (List [Char 'a', Char 'b', Char 'a'])))
      test (runPe "run")
      test (runPe "run'")
